"""
Unittest setup
"""
import contextlib
import io
import pathlib
import time
from unittest import TestCase, mock

import boto3
import responses
from botocore import UNSIGNED
from botocore.client import Config
from botocore.response import StreamingBody
from botocore.stub import Stubber

import quilt3
from quilt3.util import CONFIG_PATH


class QuiltTestCase(TestCase):
    """
    Base class for unittests.
    - Creates a mock config
    - Creates a test client
    - Mocks requests
    """
    def setUp(self):
        # Verify that CONFIG_PATH is in the test dir (patched by conftest.py).
        assert 'pytest' in str(CONFIG_PATH)

        quilt3.config(
            navigator_url='https://example.com',
            apiGatewayEndpoint='https://xyz.execute-api.us-east-1.amazonaws.com/prod',
            binaryApiGatewayEndpoint='https://xyz.execute-api.us-east-1.amazonaws.com/prod',
            default_local_registry=pathlib.Path('.').resolve().as_uri() + '/local_registry',
            default_remote_registry='s3://example/',
            default_install_location=None,
            defaultBucket='test-bucket',
            registryUrl='https://registry.example.com',
            s3Proxy='open-s3-proxy.quiltdata.com'
        )

        self.requests_mock = responses.RequestsMock(assert_all_requests_are_fired=False)
        self.requests_mock.start()

        # Create a dummy S3 client that (hopefully) can't do anything.
        boto_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
        self.s3_client = boto_client

        self.s3_client_patcher = mock.patch.multiple(
            'quilt3.data_transfer.S3ClientProvider',
            standard_client=boto_client,
            find_correct_client=lambda *args, **kwargs: boto_client,
        )
        self.s3_client_patcher.start()

        self.s3_stubber = Stubber(self.s3_client)
        self.s3_stubber.activate()

    def tearDown(self):
        self.s3_stubber.assert_no_pending_responses()
        self.s3_stubber.deactivate()
        self.s3_client_patcher.stop()
        self.requests_mock.stop()

    def s3_streaming_body(self, data):
        return StreamingBody(io.BytesIO(data), len(data))

    @contextlib.contextmanager
    def s3_test_multi_thread_download(self, bucket, key, data, *, threshold, chunksize):
        """
        Helper for testing multi-thread download of a single file.

        data is either a bytes object if a single-request download is expected,
        or a mapping like this:
        {
            'bytes=0-4': b'part1',
            'bytes=5-9': b'part2',
            ...
        }
        """
        is_single_request = isinstance(data, bytes)
        num_parts = 1 if is_single_request else len(data)
        expected_params = {
            'Bucket': bucket,
            'Key': key,
        }

        def side_effect(*args, **kwargs):
            body = self.s3_streaming_body(data if is_single_request else data[kwargs['Range']])
            if not is_single_request:
                # This ensures that we have concurrent calls to get_object().
                time.sleep(0.1 * (1 - list(data).index(kwargs['Range']) / len(data)))
            return {
                'VersionId': 'v1',
                'Body': body,
            }

        with mock.patch('quilt3.data_transfer.s3_transfer_config.multipart_threshold', threshold), \
             mock.patch('quilt3.data_transfer.s3_transfer_config.multipart_chunksize', chunksize), \
             mock.patch.object(self.s3_client, 'get_object', side_effect=side_effect) as get_object_mock:
            yield

            if is_single_request:
                get_object_mock.assert_called_once_with(**expected_params)
            else:
                assert get_object_mock.call_count == num_parts
                get_object_mock.assert_has_calls([
                    mock.call(**expected_params, Range=r)
                    for r in data
                ], any_order=True)



import textwrap

import pytest
from tests.utils import QuiltTestCase

from quilt3 import workflows
from quilt3.backends import get_package_registry
from quilt3.data_transfer import put_bytes
from quilt3.util import PhysicalKey, QuiltException


def get_v1_conf_data(conf_data):
    conf_data = textwrap.dedent(conf_data)
    return f'version: "1"\n{conf_data}'


def set_local_conf_data(conf_data):
    put_bytes(conf_data.encode(), get_package_registry().workflow_conf_pk)


def create_local_tmp_schema(data):
    pk = get_package_registry().root.join('schemas/schema')
    put_bytes(data.encode(), pk)
    return pk


class WorkflowTest(QuiltTestCase):
    def _validate(self, registry=None, workflow=..., meta=None, message=None):
        registry = get_package_registry(registry)
        meta = meta or {}
        return workflows.validate(registry=registry, workflow=workflow, meta=meta, message=message)

    def s3_mock_config(self, data, pkg_registry):
        self.s3_stubber.add_response(
            method='get_object',
            service_response={
                'VersionId': 'some-version',
                'Body': self.s3_streaming_body(data.encode()),
            },
            expected_params={
                'Bucket': pkg_registry.root.bucket,
                'Key': pkg_registry.workflow_conf_pk.path,
            }
        )

    def test_no_conf_workflow_not_specified(self):
        assert self._validate() is None

    def test_no_conf_workflow_not_specified_s3(self):
        bucket = 'some-bucket'
        self.s3_stubber.add_client_error(
            'get_object',
            service_error_code='NoSuchKey',
            expected_params={
                'Bucket': bucket,
                'Key': '.quilt/workflows/config.yml',
            },
            http_status_code=404,
        )
        assert self._validate(registry=f's3://{bucket}') is None

    def test_no_conf_workflow_specified(self):
        for workflow in (None, 'some-string'):
            with self.subTest(workflow=workflow):
                with pytest.raises(QuiltException) as excinfo:
                    self._validate(workflow=workflow)
                assert repr(workflow) in str(excinfo.value)
                assert "no workflows config exist" in str(excinfo.value)

    def test_conf_invalid_yaml(self):
        set_local_conf_data(',')
        for workflow in (None, 'some-string', ...):
            with self.subTest(workflow=workflow):
                with pytest.raises(QuiltException, match=r"Couldn't parse workflows config as YAML."):
                    self._validate(workflow=workflow)

    def test_conf_load_error_s3(self):
        bucket = 'some-bucket'
        self.s3_stubber.add_client_error(
            'get_object',
            service_error_code='AccessDenied',
            expected_params={
                'Bucket': bucket,
                'Key': '.quilt/workflows/config.yml',
            },
            http_status_code=403,
        )
        with pytest.raises(QuiltException, match=r"Couldn't load workflows config."):
            self._validate(registry=f's3://{bucket}')

    def test_conf_invalid(self):
        set_local_conf_data('')
        err_msg = r"Workflows config failed validation: None is not of type 'object'."
        for workflow in (None, 'some-string', ...):
            with self.subTest(workflow=workflow):
                with pytest.raises(QuiltException, match=err_msg):
                    self._validate(workflow=workflow)

    def test_workflow_is_required_not_specified(self):
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
        '''))
        for workflow in (None, ...):
            with self.subTest(workflow=workflow):
                with pytest.raises(QuiltException, match=r'Workflow required, but none specified.'):
                    self._validate(workflow=workflow)

    def test_workflow_is_required_default_set(self):
        set_local_conf_data(get_v1_conf_data('''
            default_workflow: w1
            workflows:
              w1:
                name: Name
        '''))

        assert self._validate() == {
            'id': 'w1',
            'config': str(get_package_registry().workflow_conf_pk),
        }

        with pytest.raises(QuiltException, match=r'Workflow required, but none specified.'):
            self._validate(workflow=None)

    def test_workflow_not_required_not_specified(self):
        set_local_conf_data(get_v1_conf_data('''
            is_workflow_required: false
            workflows:
              w1:
                name: Name
        '''))
        for workflow in (None, ...):
            with self.subTest(workflow=workflow):
                assert self._validate(workflow=workflow) == {
                    'id': None,
                    'config': str(get_package_registry().workflow_conf_pk),
                }

    def test_workflow_not_required_default_set(self):
        set_local_conf_data(get_v1_conf_data('''
            is_workflow_required: false
            default_workflow: w1
            workflows:
              w1:
                name: Name
        '''))

        assert self._validate() == {
            'id': 'w1',
            'config': str(get_package_registry().workflow_conf_pk),
        }

        assert self._validate(workflow=None) == {
            'id': None,
            'config': str(get_package_registry().workflow_conf_pk),
        }

    def test_missing_workflow(self):
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
        '''))

        with pytest.raises(QuiltException, match=r"There is no 'w2' workflow in config."):
            self._validate(workflow='w2')

    def test_missing_schema(self):
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                metadata_schema: schema-id
        '''))

        with pytest.raises(QuiltException, match=r"There is no 'schema-id' in schemas."):
            self._validate(workflow='w1')

    def test_schema_invalid_meta_schema(self):
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                metadata_schema: schema-id
            schemas:
              schema-id:
                url: %s
        ''' % create_local_tmp_schema('{"$schema": 42}')))
        with pytest.raises(QuiltException, match=r'\$schema must be a string.'):
            self._validate(workflow='w1')

    def test_schema_invalid_json(self):
        tmp_schema = create_local_tmp_schema('"')
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                metadata_schema: schema-id
            schemas:
              schema-id:
                url: %s
        ''' % tmp_schema))
        with pytest.raises(QuiltException, match=fr"Couldn't parse {tmp_schema} as JSON."):
            self._validate(workflow='w1')

    def test_schema_load_error(self):
        schema_pk = get_package_registry().root.join('nonexistent-schema')
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                metadata_schema: schema-id
            schemas:
              schema-id:
                url: %s
        ''' % schema_pk))
        with pytest.raises(QuiltException, match=fr"Couldn't load schema at {schema_pk}"):
            self._validate(workflow='w1')

    def test_schema_load_error_s3(self):
        schema_pk = PhysicalKey.from_url('s3://schema-bucket/schema-key')
        data = get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                metadata_schema: schema-id
            schemas:
              schema-id:
                url: %s
        ''' % schema_pk)
        registry = get_package_registry('s3://some-bucket')
        self.s3_mock_config(data, registry)
        self.s3_stubber.add_client_error(
            'get_object',
            service_error_code='NoSuchKey',
            expected_params={
                'Bucket': 'schema-bucket',
                'Key': 'schema-key',
            },
            http_status_code=404,
        )
        with pytest.raises(QuiltException, match=fr"Couldn't load schema at {schema_pk}"):
            self._validate(registry=registry, workflow='w1')

    def test_schema_with_ref(self):
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                metadata_schema: schema-id
            schemas:
              schema-id:
                url: %s
        ''' % create_local_tmp_schema('{"$ref": "other-schema"}')))

        with pytest.raises(QuiltException, match=r"Currently we don't support \$ref in schema."):
            self._validate(workflow='w1')

    def test_schema_validation_invalid_meta(self):
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                metadata_schema: schema-id
            schemas:
              schema-id:
                url: %s
        ''' % create_local_tmp_schema('{"type": "string"}')))

        with pytest.raises(QuiltException, match=r"Metadata failed validation: {} is not of type 'string'."):
            self._validate(workflow='w1', meta={})

    def test_schema_validation_valid_meta(self):
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                metadata_schema: schema-id
            schemas:
              schema-id:
                url: %s
        ''' % create_local_tmp_schema('{"type": "string"}')))

        assert self._validate(workflow='w1', meta="some-data") == {
            'id': 'w1',
            'config': str(get_package_registry().workflow_conf_pk),
            'schemas': {
                'schema-id': str(get_package_registry().root.join('schemas/schema')),
            },
        }

    def test_schema_validation_valid_meta_s3(self):
        schema_urls = {
            's3://schema-bucket/schema-key?versionId=schema-version': {
                'Bucket': 'schema-bucket',
                'Key': 'schema-key',
                'VersionId': 'schema-version',
            },
            's3://schema-bucket/schema-key': {
                'Bucket': 'schema-bucket',
                'Key': 'schema-key',
            },
        }
        for schema_url, expected_params in schema_urls.items():
            data = get_v1_conf_data('''
                workflows:
                  w1:
                    name: Name
                    metadata_schema: schema-id
                schemas:
                  schema-id:
                    url: %s
            ''' % schema_url)
            with self.subTest(schema_url=schema_url):
                self.s3_mock_config(data, get_package_registry('s3://some-bucket'))
                self.s3_stubber.add_response(
                    method='get_object',
                    service_response={
                        'VersionId': 'schema-version',
                        'Body': self.s3_streaming_body(b'{"type": "string"}'),
                    },
                    expected_params=expected_params,
                )
                assert self._validate(registry='s3://some-bucket', workflow='w1', meta="some-data") == {
                    'id': 'w1',
                    'config': 's3://some-bucket/.quilt/workflows/config.yml?versionId=some-version',
                    'schemas': {
                        'schema-id': 's3://schema-bucket/schema-key?versionId=schema-version',
                    },
                }

    def test_remote_registry_local_schema(self):
        data = get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                metadata_schema: schema-id
            schemas:
              schema-id:
                url: file:///local/path
        ''')
        registry = 's3://some-bucket'
        self.s3_mock_config(data, get_package_registry(registry))
        schema_pk = PhysicalKey.from_path('/local/path')
        error_msg = rf"Local schema '{schema_pk}' can't be used on the remote registry."
        with pytest.raises(QuiltException, match=error_msg):
            self._validate(registry=registry, workflow='w1')

    def test_is_message_required(self):
        set_local_conf_data(get_v1_conf_data('''
            workflows:
              w1:
                name: Name
                is_message_required: true
        '''))

        assert self._validate(workflow='w1', message='some message') == {
            'id': 'w1',
            'config': str(get_package_registry().workflow_conf_pk),
        }

        error_msg = r'Commit message is required by workflow, but none was provided.'
        for message in (None, ''):
            with self.subTest(message=message):
                with pytest.raises(QuiltException, match=error_msg):
                    self._validate(workflow='w1')

    def test_invalid_url(self):
        for url in (',', 'http://example.com', 's3://'):
            with self.subTest(url=url):
                set_local_conf_data(get_v1_conf_data(f'''
                    workflows:
                      w1:
                        name: Name
                        metadata_schema: schema-id
                    schemas:
                      schema-id:
                        url: "{url}"
                '''))
                with pytest.raises(QuiltException, match=fr"Couldn't parse URL '{url}'."):
                    self._validate(workflow='w1')

    def test_unsupported_meta_schema(self):
        for meta_schema in (
            'http://json-schema.org/draft-07/schema',
            'http://json-schema.org/draft-06/schema#',
        ):
            with self.subTest(meta_schema=meta_schema):
                set_local_conf_data(get_v1_conf_data('''
                    workflows:
                      w1:
                        name: Name
                        metadata_schema: schema-id
                    schemas:
                      schema-id:
                        url: %s
                ''' % create_local_tmp_schema(f'{{"$schema": "{meta_schema}"}}')))
                with pytest.raises(QuiltException, match=fr"Unsupported meta-schema: {meta_schema}."):
                    self._validate(workflow='w1')

    def test_supported_meta_schema(self):
        for meta_schema in (
            'http://json-schema.org/draft-07/schema#',
        ):
            with self.subTest(meta_schema=meta_schema):
                set_local_conf_data(get_v1_conf_data('''
                    workflows:
                      w1:
                        name: Name
                        metadata_schema: schema-id
                    schemas:
                      schema-id:
                        url: %s
                ''' % create_local_tmp_schema(f'{{"$schema": "{meta_schema}"}}')))
                assert self._validate(workflow='w1') == {
                    'id': 'w1',
                    'config': str(get_package_registry().workflow_conf_pk),
                    'schemas': {
                        'schema-id': str(get_package_registry().root.join('schemas/schema')),
                    },
                }



import tempfile
from pathlib import Path
from unittest import mock
from unittest.mock import patch

import pytest

import quilt3
from quilt3 import main

from .utils import QuiltTestCase

create_parser = main.create_parser


class CommandLineTestCase(QuiltTestCase):
    """
    Base TestCase class, sets up a CLI parser
    """
    @classmethod
    def setUpClass(cls):
        parser = create_parser()
        cls.parser = parser


class QuiltCLITestCase(CommandLineTestCase):
    def test_quilt_config(self):
        args = self.parser.parse_args(['config', 'https://foo.bar'])
        assert args.catalog_url == 'https://foo.bar'

    def test_push(self):
        name = 'test/name'
        pkg = quilt3.Package()

        with tempfile.TemporaryDirectory() as tmp_dir:
            (Path(tmp_dir) / 'foo').touch()
            (Path(tmp_dir) / 'bar').mkdir()
            (Path(tmp_dir) / 'bar' / 'baz')

            with mock.patch('quilt3.Package.__new__', return_value=pkg) as mocked_package_class, \
                 mock.patch.object(pkg, 'set_dir', wraps=pkg.set_dir) as mocked_set_dir, \
                 mock.patch.object(pkg, 'push') as mocked_push:
                main.main(('push', '--dir', tmp_dir, name))

                mocked_package_class.assert_called_once_with(quilt3.Package)
                mocked_set_dir.assert_called_once_with('.', tmp_dir, meta=None)
                mocked_push.assert_called_once_with(name, registry=None, dest=None, message=None)


@pytest.mark.parametrize(
    'meta_arg, meta_data, expected_set_dir_count, expected_push_count, expected_meta, expected_stderr',
    [
        (None, None, 1, 1, None, ''),
        ('--meta', '{invalid: meta}', 0, 0, {}, 'is not a valid json string'),
        ('--meta', "{'single': 'quotation'}", 0, 0, {}, 'is not a valid json string'),
        ('--meta', '{"test": "meta", }', 0, 0, {}, 'is not a valid json string'),
        ('--meta', '{"test": "meta"}', 1, 1, {"test": "meta"}, ''),
    ]
)
def test_push_with_meta_data(
    meta_arg,
    meta_data,
    expected_set_dir_count,
    expected_push_count,
    expected_meta,
    expected_stderr,
    capsys
):
    name = 'test/name'
    pkg = quilt3.Package()

    with tempfile.TemporaryDirectory() as tmp_dir:
        (Path(tmp_dir) / 'foo').touch()
        (Path(tmp_dir) / 'bar').mkdir()
        (Path(tmp_dir) / 'bar' / 'baz')

        with mock.patch('quilt3.Package.__new__', return_value=pkg) as mocked_package_class,\
             mock.patch.object(pkg, 'set_dir', wraps=pkg.set_dir) as mocked_set_dir, \
             mock.patch.object(pkg, 'push') as mocked_push:

            # '--registry' defaults to configured remote registry hence optional.
            if meta_arg:
                main.main(('push', '--dir', tmp_dir, name, meta_arg, meta_data))
            else:
                main.main(('push', '--dir', tmp_dir, name))
            mocked_package_class.assert_called_once_with(quilt3.Package)
            assert mocked_set_dir.call_count == expected_set_dir_count
            assert mocked_push.call_count == expected_push_count
            assert pkg.meta == expected_meta
            # check for expected stderr exception message
            captured = capsys.readouterr()
            assert expected_stderr in captured.err


def test_list_packages(capsys):
    registry = 's3://my_test_bucket/'
    pkg_names = ['foo/bar', 'foo/bar1', 'foo1/bar']
    with patch('quilt3.backends.s3.S3PackageRegistryV1.list_packages') as list_packages_mock:
        list_packages_mock.return_value = pkg_names
        main.main(('list-packages', registry))

        list_packages_mock.assert_called_once_with()
        captured = capsys.readouterr()
        assert captured.out.split() == pkg_names



import functools
import json

import botocore.exceptions
import jsonschema
import pkg_resources
import yaml

from quilt3.data_transfer import get_bytes_and_effective_pk

from .. import util
from ..backends import PackageRegistry


@functools.lru_cache(maxsize=None)
def _get_conf_validator():
    schema = json.loads(pkg_resources.resource_string(__name__, 'config-1.schema.json'))
    return jsonschema.Draft7Validator(schema).validate


SUPPORTED_META_SCHEMAS = {
    'http://json-schema.org/draft-07/schema#': jsonschema.Draft7Validator,
}


def _schema_load_object_hook(o):
    if '$ref' in o:
        raise util.QuiltException("Currently we don't support $ref in schema.")
    return o


_load_schema_json = json.JSONDecoder(object_hook=_schema_load_object_hook).decode


def validate(*, registry: PackageRegistry, workflow, meta, message):
    # workflow is ... => no workflow provided by user;
    # workflow is None => don't use any workflow.
    if not (workflow in (None, ...) or isinstance(workflow, str)):
        raise TypeError

    conf_info = None
    try:
        conf_info = get_bytes_and_effective_pk(registry.workflow_conf_pk)
    except FileNotFoundError:
        pass
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] != 'NoSuchKey':
            raise util.QuiltException(f"Couldn't load workflows config. {e}.")
    if conf_info is None:
        if workflow is ...:
            return
        raise util.QuiltException(f'{workflow!r} workflow is specified, but no workflows config exist.')
    conf_data, conf_pk = conf_info
    try:
        # TODO: raise if objects contain duplicate properties
        conf_data = yaml.safe_load(conf_data.decode())
    except yaml.YAMLError as e:
        raise util.QuiltException("Couldn't parse workflows config as YAML.") from e
    conf_validator = _get_conf_validator()
    try:
        conf_validator(conf_data)
    except jsonschema.ValidationError as e:
        raise util.QuiltException(f'Workflows config failed validation: {e.message}.') from e

    if workflow is ...:
        workflow = conf_data.get('default_workflow')

    result = {
        'id': workflow,
        'config': str(conf_pk),
    }
    if workflow is None:
        if conf_data.get('is_workflow_required', True):
            raise util.QuiltException('Workflow required, but none specified.')
        return result

    workflows_data = conf_data['workflows']
    if workflow not in workflows_data:
        raise util.QuiltException(f'There is no {workflow!r} workflow in config.')
    workflow_data = workflows_data[workflow]
    metadata_schema_id = workflow_data.get('metadata_schema')
    if metadata_schema_id:
        schemas = conf_data.get('schemas', {})
        if metadata_schema_id not in schemas:
            raise util.QuiltException(f'There is no {metadata_schema_id!r} in schemas.')
        schema_url = schemas[metadata_schema_id]['url']
        try:
            schema_pk = util.PhysicalKey.from_url(schema_url)
        except util.URLParseError as e:
            raise util.QuiltException(f"Couldn't parse URL {schema_url!r}. {e}.")
        if schema_pk.is_local() and not registry.is_local:
            raise util.QuiltException(f"Local schema {str(schema_pk)!r} can't be used on the remote registry.")

        handled_exception = (OSError if schema_pk.is_local() else botocore.exceptions.ClientError)
        try:
            schema_data, schema_pk_to_store = get_bytes_and_effective_pk(schema_pk)
        except handled_exception as e:
            raise util.QuiltException(f"Couldn't load schema at {schema_pk}. {e}.")
        try:
            schema = _load_schema_json(schema_data.decode())
        except json.JSONDecodeError as e:
            raise util.QuiltException(f"Couldn't parse {schema_pk} as JSON. {e}.")

        validator_cls = jsonschema.Draft7Validator
        if isinstance(schema, dict) and '$schema' in schema:
            meta_schema = schema['$schema']
            if not isinstance(meta_schema, str):
                raise util.QuiltException('$schema must be a string.')
            validator_cls = SUPPORTED_META_SCHEMAS.get(meta_schema)
            if validator_cls is None:
                raise util.QuiltException(f"Unsupported meta-schema: {meta_schema}.")
        try:
            jsonschema.validate(meta, schema, cls=validator_cls)
        except jsonschema.ValidationError as e:
            raise util.QuiltException(f"Metadata failed validation: {e.message}.")
        result['schemas'] = {metadata_schema_id: str(schema_pk_to_store)}
    if workflow_data.get('is_message_required', False) and not message:
        raise util.QuiltException('Commit message is required by workflow, but none was provided.')

    return result
